from contextlib import suppress
import importlib
from pathlib import Path
from typing import Any, Mapping, TypeVar, TYPE_CHECKING
import uuid
import warnings
from pydantic import BaseModel, Field

from himena._descriptors import NoNeedToSave, SaveToPath
from himena._utils import get_widget_class_id
from himena.layout import construct_layout
from himena.types import WindowState, WindowRect, WidgetDataModel
from himena.utils.misc import is_subtype
from himena import anchor
from himena._providers import ReaderStore
from himena.widgets._wrapper import ParametricWindow
from himena.standards import read_metadata
from himena.workflow import Workflow, compute, WorkflowStepType, LocalReaderMethod

if TYPE_CHECKING:
    from himena.widgets import SubWindow, TabArea, MainWindow

_W = TypeVar("_W")  # widget interface


class WindowDescription(BaseModel):
    """A model that describes a window state."""

    title: str
    rect: WindowRect
    state: WindowState = Field(default=WindowState.NORMAL)
    anchor: dict[str, Any] = Field(default_factory=lambda: {"type": "no-anchor"})
    is_editable: bool = Field(default=True)
    id: uuid.UUID
    short_workflow: WorkflowStepType | None = Field(default=None)
    workflow: Workflow
    model_type: str
    widget_plugin_id: str | None = Field(default=None)
    children: list[uuid.UUID] = Field(default_factory=list)

    @classmethod
    def from_gui(
        cls,
        window: "SubWindow",
        *,
        allow_calculate: bool = False,
    ) -> "WindowDescription":
        """Construct a WindowDescription from a SubWindow instance."""
        read_from = window._determine_read_from()
        if read_from is None:
            if allow_calculate:
                short_workflow = None
            else:
                raise ValueError("Cannot determine where to read the model from.")
        else:
            short_workflow = LocalReaderMethod(
                path=read_from[0],
                plugin=read_from[1],
            )
        model = window.to_model()
        return WindowDescription(
            title=window.title,
            rect=window.rect,
            state=window.state,
            anchor=anchor.anchor_to_dict(window.anchor),
            is_editable=window.is_editable,
            id=window._identifier,
            short_workflow=short_workflow,
            workflow=model.workflow,
            model_type=model.type,
            widget_plugin_id=get_widget_class_id(type(window.widget)),
            children=[child._identifier for child in window._child_windows],
        )

    def process_model(self, area: "TabArea[_W]", model: "WidgetDataModel"):
        """Add model to the tab area and update the window properties."""
        model.workflow = self.workflow
        model.save_behavior_override = NoNeedToSave()
        if widget_plugin_id := self.widget_plugin_id:
            try:
                window = area.add_data_model(model.with_open_plugin(widget_plugin_id))
            except Exception as e:
                window = area.add_data_model(model)
                warnings.warn(
                    "Could not open the file with the intended widget plugin "
                    f"{widget_plugin_id} because of the following error: {e}",
                    RuntimeWarning,
                    stacklevel=2,
                )
        else:
            window = area.add_data_model(model)
        window.title = self.title
        window.rect = self.rect
        window.state = self.state
        window.anchor = anchor.dict_to_anchor(self.anchor)
        with suppress(AttributeError):
            window.is_editable = self.is_editable
        if isinstance(meth := self.short_workflow, LocalReaderMethod):
            window._save_behavior = SaveToPath(path=meth.path, plugin=meth.plugin)
        window._identifier = self.id
        return window

    def prep_workflow(
        self, workflow_override: Mapping[uuid.UUID, Workflow]
    ) -> Workflow:
        """Prepare the most efficient workflow to get the window."""
        if wf := workflow_override.get(self.id):
            pass
        else:
            if self.short_workflow is None:
                wf = self.workflow
            else:
                wf = Workflow(steps=[self.short_workflow])
        if len(wf) == 1 and isinstance(meth := wf[0], LocalReaderMethod):
            # look for the best reader according to _win_sessions.model_type
            if meth.plugin is None:
                meth.plugin = _pick_best_reader_plugin(meth, self.model_type)
        return wf


class TabSession(BaseModel):
    """A session of a tab."""

    name: str = Field(default="")
    windows: list[WindowDescription] = Field(default_factory=list)
    current_index: int | None = Field(default=None)
    layouts: list[dict] = Field(default_factory=list)

    @classmethod
    def from_gui(
        cls,
        tab: "TabArea[_W]",
        *,
        allow_calculate: bool = False,
    ) -> "TabSession":
        layouts: list[dict] = []
        for ly in tab.layouts:
            if ly is tab._minimized_window_stack_layout:
                continue  # this layout does not need to be saved
            layouts.append(ly._serialize_layout())
        return TabSession(
            name=tab.name,
            windows=[
                WindowDescription.from_gui(window, allow_calculate=allow_calculate)
                for window in tab
                if not isinstance(window, ParametricWindow)
            ],
            current_index=tab.current_index,
            layouts=layouts,
        )

    def update_gui(
        self,
        main: "MainWindow[_W]",
        *,
        workflow_override: Mapping[str, Workflow] = {},
        dirpath: Path,
    ) -> None:
        """Update the GUI state based on the session."""
        _win_sessions: list[tuple[int, WindowDescription]] = []
        _pending_workflows: list[Workflow] = []
        area = main.add_tab(self.name)
        cur_index = self.current_index
        for i_win, window_session in enumerate(self.windows):
            _win_sessions.append((i_win, window_session))
            wf = window_session.prep_workflow(workflow_override)
            _pending_workflows.append(wf)

        models = compute(_pending_workflows)
        _failed_sessions: list[tuple[WindowDescription, Exception]] = []
        _id_to_win: dict[uuid.UUID, "SubWindow"] = {}
        for (i_win_sess, _win_sess), model_or_exc in zip(_win_sessions, models):
            if isinstance(model_or_exc, Exception):
                _failed_sessions.append((_win_sess, model_or_exc))
                continue
            # look for the metadata
            meta_path = dirpath / f"{i_win_sess}_{_win_sess.title}.himena-meta"
            if meta_path.exists():
                try:
                    model_or_exc.metadata = read_metadata(meta_path)
                except Exception as e:
                    warnings.warn(
                        f"Failed to read metadata from {meta_path}: {e}",
                        RuntimeWarning,
                        stacklevel=2,
                    )
            _win = _win_sess.process_model(self, model_or_exc)
            _id_to_win[_win_sess.id] = _win

        if 0 <= cur_index < len(area):
            area.current_index = cur_index
        _raise_failed(_failed_sessions)
        _update_layout(area, self.layouts, main)
        for _win_sess in self.windows:
            for child_id in _win_sess.children:
                _id_to_win[_win_sess.id]._child_windows.add(_id_to_win[child_id])
        return None


def _update_layout(tab: "TabArea[_W]", layouts: list[dict], main: "MainWindow[_W]"):
    for ly in layouts:
        _layout_obj = construct_layout(ly, main)
        tab._add_layout_impl(_layout_obj)
    return None


def _get_version(mod, maybe_file: bool = False) -> str | None:
    if maybe_file and Path(mod).suffix:
        return None
    if isinstance(mod, str):
        mod = importlib.import_module(mod)
    return getattr(mod, "__version__", None)


class AppProfileInfo(BaseModel):
    """A simplified app profile for saving a session."""

    name: str
    plugins: list[str] = Field(default_factory=list)
    theme: str


class AppSession(BaseModel):
    """A session of the entire application."""

    version: str | None = Field(default_factory=lambda: _get_version("himena"))
    profile: AppProfileInfo | None = Field(default=None)
    tabs: list[TabSession] = Field(default_factory=list)
    current_index: int = Field(default=0)
    rect: WindowRect = Field(default=WindowRect(200, 200, 800, 600))

    @classmethod
    def from_gui(
        cls,
        main: "MainWindow[_W]",
        *,
        allow_calculate: bool = False,
    ) -> "AppSession":
        app_prof = main.app_profile
        profile = AppProfileInfo(
            name=app_prof.name,
            plugins=app_prof.plugins,
            theme=app_prof.theme,
        )
        return AppSession(
            profile=profile,
            tabs=[
                TabSession.from_gui(tab, allow_calculate=allow_calculate)
                for tab in main.tabs
            ],
            current_index=main.tabs.current_index,
            rect=main.rect,
        )

    def update_gui(
        self,
        main: "MainWindow[_W]",
        *,
        workflow_override: Mapping[str, Workflow] = {},
        dirpath: Path,
    ) -> None:
        """Update the GUI state based on the session."""
        cur_index = self.current_index
        _tab_sessions: list[tuple[int, TabSession]] = []
        _win_sessions: list[tuple[int, WindowDescription]] = []
        _target_areas: list[tuple[int, TabArea]] = []
        _pending_workflows: list[Workflow] = []
        for i_tab, tab_session in enumerate(self.tabs):
            _tab_sessions.append((i_tab, tab_session))
            _new_tab = main.add_tab(tab_session.name)
            for i_win, window_session in enumerate(tab_session.windows):
                _win_sessions.append((i_win, window_session))
                wf = window_session.prep_workflow(workflow_override)
                _pending_workflows.append(wf)
                _target_areas.append((i_tab, _new_tab))

        # read all the metadata
        all_metadata = []
        for (i_win, _win_sess), (i_tab, _tab_area) in zip(_win_sessions, _target_areas):
            # look for the metadata
            meta = None
            meta_path = (
                dirpath
                / f"{i_tab}_{_tab_area.title}"
                / f"{i_win}_{_win_sess.title}.himena-meta"
            )
            if meta_path.exists():
                try:
                    meta = read_metadata(meta_path)
                except Exception as e:
                    warnings.warn(
                        f"Failed to read metadata from {meta_path}: {e}",
                        RuntimeWarning,
                        stacklevel=2,
                    )
            all_metadata.append(meta)
        models = compute(_pending_workflows, all_metadata)
        _failed_sessions: list[tuple[WindowDescription, Exception]] = []
        _id_to_win: dict[uuid.UUID, "SubWindow"] = {}
        for (i_win, _win_sess), (i_tab, _tab_area), model_or_exc in zip(
            _win_sessions, _target_areas, models
        ):
            if isinstance(model_or_exc, Exception):
                _failed_sessions.append((_win_sess, model_or_exc))
                continue
            _id_to_win[_win_sess.id] = _win_sess.process_model(_tab_area, model_or_exc)

        # Update current active window for each tab
        for (_, tab_session), (_, area) in zip(_tab_sessions, _target_areas):
            cur_tab_index = tab_session.current_index
            if cur_tab_index is not None and 0 <= cur_tab_index < len(area):
                area.current_index = cur_tab_index
        main.tabs.current_index = self.current_index + cur_index
        _raise_failed(_failed_sessions)
        main.rect = self.rect
        for tab_session in self.tabs:
            _update_layout(main.tabs[tab_session.name], tab_session.layouts, main)

        # connect window children to their parents
        for _, _win_sess in _win_sessions:
            for child_id in _win_sess.children:
                _id_to_win[_win_sess.id]._child_windows.add(_id_to_win[child_id])
        return None


def _raise_failed(failed: list[tuple[WindowDescription, Exception]]) -> None:
    if len(failed) > 0:
        msg = "Could not load the following windows:\n"
        list_of_failed = "\n".join(
            f"- {win.title} ({type(exc).__name__}: {exc})" for win, exc in failed
        )
        raise ValueError(msg + list_of_failed) from failed[-1][1]


def _pick_best_reader_plugin(meth: LocalReaderMethod, expected_type: str) -> str | None:
    ins = ReaderStore().instance()
    suboptimals: list[int, str] = []
    for model_type, reader in ins.iter_readers(meth.path, min_priority=-float("inf")):
        if model_type is None:
            continue
        if model_type == expected_type:
            return reader.plugin.to_str()
        elif is_subtype(model_type, expected_type):
            if reader.plugin is None:
                continue
            score = len(model_type)
            suboptimals.append((score, reader.plugin.to_str()))
    if len(suboptimals) == 0:
        return None
    return max(suboptimals, key=lambda x: x[0])[1]



from __future__ import annotations

from qtpy import QtCore, QtWidgets as QtW, QtGui
from himena.utils.enum import StrEnum


class ToolTipBehavior(StrEnum):
    """Tooltip behavior when the parent window loses focus."""

    STAY = "stay"  # Stay in place
    FOLLOW = "follow"  # Follow the cursor
    UNTIL_MOVE = "until_move"  # Stay until the cursor moves


class QToolTipWidget(QtW.QLabel):
    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent, QtCore.Qt.WindowType.ToolTip)
        self.setObjectName("HimenaToolTip")
        self.setWindowFlag(QtCore.Qt.WindowType.FramelessWindowHint)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.hide()
        self._timer = QtCore.QTimer(self)
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self.hide)
        self._offset = QtCore.QPoint(14, 18)
        self._behavior = ToolTipBehavior.FOLLOW

    def set_behavior(self, behavior: str) -> None:
        self._behavior = ToolTipBehavior(behavior)

    def show_tooltip(self, text: str, duration: float = 3.0) -> None:
        """Show tooltip next to the cursor for a duration (sec)."""
        self._timer.stop()
        if not text:
            self.setText(text)
            self.hide()
        else:
            self.setText(text)
            self.adjustSize()
            self.move_tooltip(QtGui.QCursor.pos())
            self.show()
            self._timer.start(int(duration * 1000))

    # def show_image(self, image: QtGui.QImage, duration: float = 3.0) -> None:
    #     """Show an image as tooltip next to the cursor."""
    #     self._timer.stop()
    #     if image is None:
    #         self.clear()
    #         self.hide()
    #     else:
    #         self.setPixmap(QtGui.QPixmap.fromImage(image))
    #         self.adjustSize()
    #         self.move_tooltip(QtGui.QCursor.pos())
    #         self.show()
    #         self._timer.start(int(duration * 1000))

    def move_tooltip(self, pos: QtCore.QPoint) -> None:
        self.move(pos + self._offset)



import numpy as np
from numpy.testing import assert_equal
import pytest
from pytestqt.qtbot import QtBot
from qtpy.QtCore import Qt
import pandas as pd
import polars as pl
from himena import MainWindow
from himena.consts import StandardType
from himena.core import create_model
from himena.testing.subwindow import WidgetTester
from himena.workflow import CommandExecution
from himena_builtins.qt.dataframe import QDataFrameView, QDataFramePlotView
from himena_builtins.qt.widgets.dataframe import select_columns
from himena_builtins.qt.basic import QDictView

_Ctrl = Qt.KeyboardModifier.ControlModifier

@pytest.mark.parametrize(
    "df",
    [
        {"a": [1, -2], "b": [3.0, -4.0], "str": ["a", "b"]},
        {"a": [3], "b": ["ggg"]},
        pd.DataFrame({"a": [1, -2], "b": [3.0, -4.0], "str": ["a", "b"]}),
        pl.DataFrame({"a": [1, -2], "b": [3.0, -4.0], "str": ["a", "b"]}),
    ],
)
def test_dataframe(himena_ui: MainWindow, qtbot: QtBot, df):
    with WidgetTester(QDataFrameView(himena_ui)) as tester:
        tester.update_model(value=df)
        qtbot.addWidget(tester.widget)
        table = tester.widget
        table.selection_model.set_ranges([(slice(1, 2), slice(1, 3))])
        table.copy_data()
        table._make_context_menu()
        table.model().headerData(0, Qt.Orientation.Horizontal, Qt.ItemDataRole.ToolTipRole)
        table.model().headerData(0, Qt.Orientation.Vertical, Qt.ItemDataRole.ToolTipRole)
        qtbot.keyClick(tester.widget, Qt.Key.Key_F, modifier=_Ctrl)
        finder = tester.widget._finder_widget
        assert finder is not None
        finder._line_edit.setText("b")
        qtbot.keyClick(finder, Qt.Key.Key_Enter)
        qtbot.keyClick(finder, Qt.Key.Key_Enter, modifier=Qt.KeyboardModifier.ShiftModifier)
        assert type(tester.to_model().value) is type(df)
        tester.is_modified()
        assert tester.widget._hor_header._data_model_for_drag() is not None
        table.selection_model.append((slice(0, 3), slice(0, 1)), column=True)
        tester.widget._hor_header._process_move_event(0)
        tester.widget._hor_header._process_move_event(1)
        select_columns(tester.to_model())([0])

def test_dataframe_plot(himena_ui: MainWindow, qtbot: QtBot):
    x = np.linspace(0, 3, 20)
    df = {"x": x, "y": np.sin(x * 2), "z": np.cos(x * 2)}
    with WidgetTester(QDataFramePlotView(himena_ui)) as tester:
        tester.update_model(value=df)
        tester.cycle_model()
        qtbot.addWidget(tester.widget)
        tester.widget._table_widget.selection_model.move_to(0, 1)
        tester.widget._table_widget.selection_model.set_ranges([(slice(0, 20), slice(1, 2))])

def test_dataframe_command(himena_ui: MainWindow):
    win = himena_ui.add_object(
        {"a": [1, 2], "b": ["p", "q"]},
        type=StandardType.DATAFRAME,
    )
    himena_ui.exec_action("builtins:dataframe:header-to-row")
    assert isinstance(step := win.to_model().workflow.last(), CommandExecution)
    assert step.command_id == "builtins:dataframe:header-to-row"
    win = himena_ui.add_object(
        {"a": [1, 2], "b": ["p", "q"]},
        type=StandardType.DATAFRAME,
    )
    himena_ui.exec_action("builtins:dataframe:series-as-array", with_params={"column": "a"})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe:select-columns-by-name", with_params={"columns": ["b"]})
    assert _data_frame_equal(himena_ui.current_model.value, {"b": ["p", "q"]})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe:filter", with_params={"column": "b", "operator": "eq", "value": "p"})
    assert _data_frame_equal(himena_ui.current_model.value, {"a": [1], "b": ["p"]})
    himena_ui.current_window = win
    himena_ui.exec_action("builtins:dataframe:sort", with_params={"column": "b", "descending": True, "inplace": True})
    assert _data_frame_equal(himena_ui.current_model.value, {"a": [2, 1], "b": ["q", "p"]})
    assert _data_frame_equal(win.to_model().value, {"a": [2, 1], "b": ["q", "p"]})
    assert win.to_model().workflow.last().command_id == "builtins:dataframe:sort"

    win = himena_ui.current_window
    fn = create_model(
        lambda x: x + 1,
        type=StandardType.FUNCTION,
        add_empty_workflow=True,
    )
    himena_ui.exec_action(
        "builtins:dataframe:new-column-using-function",
        with_params={
            "input_column_name": "a",
            "function": fn,
            "output_column_name": "a2",
        },
        window_context=win,
    )

@pytest.mark.parametrize(
    "df",
    [
        {"a": np.array([1, -2]), "b": np.array([3.0, -4.0]), "str": np.array(["a", "b"])},
        pd.DataFrame({"a": [1, -2], "b": [3.0, -4.0], "str": ["a", "b"]}),
        pl.DataFrame({"a": [1, -2], "b": [3.0, -4.0], "str": ["a", "b"]}),
    ],
)
def test_copy_on_write(himena_ui: MainWindow, qtbot: QtBot, df):
    view = QDataFrameView(himena_ui)
    qtbot.addWidget(view)
    with WidgetTester(view) as tester:
        tester.update_model(value=df, type="dataframe")
        view.edit_item(0, 0, "100")
        assert_equal(np.array(tester.to_model().value["a"]), [100, -2])
        assert_equal(np.array(df["a"]), [1, -2])
        view2 = QDataFrameView(himena_ui)
        qtbot.addWidget(view2)
        view2.update_model(view.to_model())
        view.edit_item(1, 1, "0.1")
        assert_equal(np.array(tester.to_model().value["b"]), [3.0, 0.1])
        assert_equal(np.array(df["b"]), [3.0, -4.0])
        assert_equal(np.array(view2.to_model().value["b"]), [3.0, -4.0])

@pytest.mark.parametrize(
    "input_value",
    [
        {"a": 2, "b": "string", "cc": -1.4, "nan": np.nan, "empty": None},
        {"maybe_array": np.arange(5), "maybe_list": [1, 2, 3], "maybe_dict": {"a": 1}},
    ],
)
def test_dict_view(himena_ui: MainWindow, qtbot: QtBot, input_value):
    view = QDictView(himena_ui)
    qtbot.addWidget(view)
    with WidgetTester(view) as tester:
        tester.update_model(value=input_value, type=StandardType.DICT)
        tester.cycle_model()

def _data_frame_equal(a: dict, b: dict):
    for k in a.keys():
        if not np.all(a[k] == b[k]):
            return False
    return True



from __future__ import annotations

from typing import Any
from himena.consts import StandardType


class ObjectTypeMap:
    def __init__(self):
        self._functions = []

    def pick_type(self, value: Any) -> tuple[str, Any, Any]:
        for func in reversed(self._functions):
            out = func(value)
            if isinstance(out, tuple):
                typ, value_processed, meta = out
                if not isinstance(typ, str):
                    raise TypeError(
                        f"Object type map function {func} must return a str or a tuple "
                        f"of (str, Any, Any), got {out!r}."
                    )
            elif isinstance(out, str):
                typ, value_processed, meta = out, value, None
            elif out is None:
                continue
            else:
                raise TypeError(
                    f"Object type map function {func} must return a str or a tuple "
                    f"of (str, Any, Any), got {out!r}."
                )
            return typ, value_processed, meta
        raise ValueError(f"Could not determine the type of {value}.")

    def register(self, func):
        self._functions.append(func)
        return func


def register_defaults(map: ObjectTypeMap):
    import numpy as np
    from himena.workflow import Workflow

    @map.register
    def str_as_text(value: Any) -> str | None:
        if isinstance(value, str):
            return StandardType.TEXT
        return None

    @map.register
    def as_array(value) -> str | None:
        if isinstance(value, np.ndarray):
            if value.ndim == 2 and isinstance(value.dtype, np.dtypes.StringDType):
                return StandardType.TABLE
            return StandardType.ARRAY
        return None

    @map.register
    def as_dataframe(value) -> str | None:
        if hasattr(value, "__dataframe__"):
            return StandardType.DATAFRAME
        return None

    @map.register
    def as_workflow(value) -> str | None:
        if isinstance(value, Workflow):
            return StandardType.WORKFLOW
        return None



"""Widget wrappers, including sub-window and dock widget."""

from __future__ import annotations

from concurrent.futures import Future
from contextlib import suppress
import inspect
import logging
from pathlib import Path
from typing import Any, Callable, Generic, TYPE_CHECKING, Literal, TypeVar, overload
import uuid
import weakref

from psygnal import Signal
from magicgui import widgets as mgw
from himena import anchor as _anchor
from himena import io_utils
from himena._utils import get_widget_class_id
from himena.utils.window_rect import prevent_window_overlap
from himena.consts import ParametricWidgetProtocolNames as PWPN
from himena.types import (
    BackendInstructions,
    DragDataModel,
    DropResult,
    ModelTrack,
    Parametric,
    ParametricWidgetProtocol,
    Size,
    WindowState,
    WidgetDataModel,
    WindowRect,
    FutureInfo,
)
from himena._descriptors import NoNeedToSave, SaveBehavior, SaveToNewPath, SaveToPath
from himena.workflow import (
    CommandExecution,
    LocalReaderMethod,
    Workflow,
)
from himena.plugins import _checker, AppActionRegistry
from himena.plugins._signature import _is_annotated, _split_annotated_type
from himena.layout import Layout
from himena.widgets._modifications import Modifications

if TYPE_CHECKING:
    from himena.widgets import BackendMainWindow, MainWindow, TabArea

_W = TypeVar("_W")  # backend widget type
_LOGGER = logging.getLogger(__name__)


class _HasMainWindowRef(Generic[_W]):
    def __init__(self, main_window: BackendMainWindow[_W]):
        self._main_window_ref = weakref.ref(main_window)

    def _main_window(self) -> BackendMainWindow[_W]:
        out = self._main_window_ref()
        if out is None:
            raise RuntimeError("Main window was deleted")
        return out


class StrongRef(Generic[_W]):
    def __init__(self, obj: _W):
        self._obj = obj

    def __call__(self) -> _W:
        return self._obj


class WidgetWrapper(_HasMainWindowRef[_W]):
    def __init__(
        self,
        widget: _W,
        main_window: BackendMainWindow[_W],
    ):
        super().__init__(main_window)
        self._identifier = uuid.uuid4()
        self._save_behavior: SaveBehavior = SaveToNewPath()
        self._widget_workflow: Workflow = Workflow()
        self._ask_save_before_close = False
        interf, front = _split_interface_and_frontend(widget)
        front._himena_widget = self
        if interf is front:
            # the frontend main window will keep the widget, thus this wrapper just
            # needs a weak reference
            self._widget = weakref.ref(widget)
        else:
            # the frontend main window will not keep the widget, thus this wrapper needs
            # a strong reference
            self._widget = StrongRef(widget)
        self._extension_default_fallback: str | None = None
        self._force_not_editable: bool = False
        self._data_modifications = Modifications()

    @property
    def is_alive(self) -> bool:
        """Whether the widget is present in a main window."""
        # by default, always return True
        return True

    @property
    def widget(self) -> _W:
        """Get the internal backend widget."""
        if (out := self._widget()) is not None:
            return out
        # NOTE: do not call repr(self) because it will cause infinite recursion
        raise RuntimeError("Widget in the wrapper was deleted.")

    @property
    def save_behavior(self) -> SaveBehavior:
        """Get the save behavior of the widget."""
        return self._save_behavior

    def update_default_save_path(
        self,
        path: str | Path,
        plugin: str | None = None,
    ) -> None:
        """Update the save behavior of the widget."""
        if isinstance(self._save_behavior, SaveToPath):
            ask_overwrite = self._save_behavior.ask_overwrite
        else:
            ask_overwrite = True
        self._save_behavior = SaveToPath(
            path=Path(path), ask_overwrite=ask_overwrite, plugin=plugin
        )
        self._set_ask_save_before_close(False)
        return None

    def _update_model_workflow(
        self,
        workflow: Workflow | None,
        overwrite: bool = True,
    ) -> None:
        """Update the method descriptor of the widget."""
        if len(self._widget_workflow) == 0 or overwrite:
            self._widget_workflow = workflow or Workflow()
            if last_step := workflow.last():
                self._identifier = last_step.id
            _LOGGER.info("Workflow of %r updated to %r", self, workflow)
        else:
            _LOGGER.info(
                "Workflow of %r was not updated because old workflow is %r",
                self,
                self._widget_workflow,
            )
        return None

    @property
    def supports_update_model(self) -> bool:
        """Whether the widget interface supports being updated by a WidgetDataModel."""
        return hasattr(self.widget, "update_model")

    @property
    def supports_to_model(self) -> bool:
        """Whether the widget interface supports being converted to a WidgetDataModel."""
        return hasattr(self.widget, "to_model")

    @property
    def is_modified(self) -> bool:
        """Whether the content of the widget has been modified by user."""
        if hasattr(self.widget, "is_modified"):
            return self.widget.is_modified()
        return False

    @property
    def is_editable(self) -> bool:
        """Whether the widget is in an editable state."""
        is_editable_func = getattr(self.widget, "is_editable", None)
        return callable(is_editable_func) and is_editable_func()

    @is_editable.setter
    def is_editable(self, value: bool) -> None:
        set_editable_func = getattr(self.widget, "set_editable", None)
        if not callable(set_editable_func):
            raise AttributeError("Widget does not have `set_editable` method.")
        if self._force_not_editable and value:
            raise ValueError("Widget is forced to be not editable.")
        set_editable_func(value)

    def force_not_editable(self, force: bool):
        self._force_not_editable = force
        if force:
            with suppress(AttributeError):
                self.is_editable = False

    def _set_ask_save_before_close(self, value: bool) -> None:
        """Set the modified state of the widget."""
        if value and not self.supports_to_model:
            # If the backend widget cannot be converted to a model, there's no need
            # to inform the user "save changes?".
            return None
        if hasattr(self.widget, "set_modified") and not value:
            # this clause is needed after user chose "Overwrite" in the dialog
            self.widget.set_modified(False)
        self._ask_save_before_close = value
        return None

    def _need_ask_save_before_close(self) -> bool:
        """Whether the widget needs to ask the user to save before closing."""
        return self._ask_save_before_close or self.is_modified

    def size_hint(self) -> tuple[int, int] | None:
        """Size hint of the sub-window."""
        return getattr(self.widget, "size_hint", _do_nothing)()

    def model_type(self) -> str | None:
        """Type of the widget data model."""
        if not self.supports_to_model:
            return None
        interf = self.widget
        _type = None
        if hasattr(interf, "model_type"):
            _type = interf.model_type()
        elif hasattr(interf, "__himena_model_type__"):
            _type = interf.__himena_model_type__
        if _type is None:
            _type = self.to_model().type
        return _type

    @overload
    def update_model(self, model: WidgetDataModel) -> None: ...
    @overload
    def update_model(
        self,
        value: Any,
        *,
        type: str | None = None,
        metadata: Any | None = None,
        extension_default: str | None = None,
        extensions: list[str] = (),
    ) -> None: ...

    def update_model(self, value, **kwargs) -> None:
        """Update the widget by a widget data model."""
        if not self.supports_update_model:
            raise ValueError("Widget does not have `update_model` method.")
        if isinstance(value, WidgetDataModel):
            if kwargs:
                raise TypeError("Keyword arguments are not allowed with model input.")
            model = value
        else:
            allowed = {"type", "metadata", "extension_default", "extensions"}
            if invalid := set(kwargs.keys()) - allowed:
                raise ValueError(f"Invalid keyword arguments: {invalid!r}")
            model = WidgetDataModel(value=value, **kwargs)
        self.widget.update_model(model)

    def _is_drop_accepted(self, incoming: DragDataModel) -> bool:
        widget = self.widget
        return incoming.widget_accepts_me(widget)

    def _split_interface_and_frontend(self) -> tuple[object, _W]:
        """Split the interface that defines methods and the frontend widget.

        This function is used to separate the interface object that implements the
        himena protocols and the actual widget that will be added to the main window.
        """
        return _split_interface_and_frontend(self.widget)

    def _frontend_widget(self) -> _W:
        """Get the frontend widget."""
        return self._split_interface_and_frontend()[1]


def _split_interface_and_frontend(obj: _W) -> tuple[object, _W]:
    if hasattr(obj, "native_widget"):
        front = obj.native_widget()
    elif isinstance(obj, mgw.Widget):
        front = obj.native
    else:
        front = obj
    return obj, front


class SubWindow(WidgetWrapper[_W], Layout):
    state_changed = Signal(WindowState)
    renamed = Signal(str)
    closed = Signal()

    def __init__(
        self,
        widget: _W,
        main_window: BackendMainWindow[_W],
    ):
        super().__init__(widget, main_window=main_window)
        Layout.__init__(self, main_window)
        self._child_windows: weakref.WeakSet[SubWindow[_W]] = weakref.WeakSet()
        self._alive = False
        self.closed.connect(self._close_callback)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(title={self.title!r}, widget={_widget_repr(self)})"
        )

    def __class_getitem__(cls, widget_type: type[_W]):
        # this hack allows in_n_out to assign both SubWindow and SubWindow[T] to the
        # same provider/processor.
        return cls

    @classmethod
    def _deserialize_layout(cls, obj: dict, main: MainWindow) -> SubWindow[Any]:
        win = main.window_for_id(uuid.UUID(obj["id"]))
        if win is None:
            raise RuntimeError(f"SubWindow {obj['id']} not found in main window.")
        return win

    def _serialize_layout(self):
        return {"type": "subwindow", "id": self._identifier.hex}

    @property
    def title(self) -> str:
        """Title of the sub-window."""
        return self._main_window()._window_title(self._frontend_widget())

    @title.setter
    def title(self, value: str) -> None:
        self._main_window()._set_window_title(self._frontend_widget(), value)

    @property
    def state(self) -> WindowState:
        """State (e.g. maximized, minimized) of the sub-window."""
        return self._main_window()._window_state(self._frontend_widget())

    @state.setter
    def state(self, value: WindowState | str) -> None:
        main = self._main_window()._himena_main_window
        inst = main._instructions.updated(animate=False)
        self._set_state(value, inst)

    @property
    def rect(self) -> WindowRect:
        """Position and size of the sub-window."""
        return self._main_window()._window_rect(self._frontend_widget())

    @rect.setter
    def rect(self, value: tuple[int, int, int, int]) -> None:
        main = self._main_window()._himena_main_window
        inst = main._instructions.updated(animate=False)
        self._set_rect(value, inst)

    @property
    def is_alive(self) -> bool:
        """Whether the sub-window is present in a main window."""
        return self._alive

    def to_model(self) -> WidgetDataModel:
        """Export the widget data."""
        if not self.supports_to_model:
            raise ValueError("Widget does not have `to_model` method.")
        model = self.widget.to_model()  # type: ignore
        if not isinstance(model, WidgetDataModel):
            raise TypeError(
                "`to_model` method must return an instance of WidgetDataModel, got "
                f"{type(model)}"
            )

        if model.title is None:
            model.title = self.title
        if len(model.workflow) == 0:
            model.workflow = self._widget_workflow
        if self.is_modified:
            self._data_modifications.update_workflow(model)
        if model.extension_default is None:
            model.extension_default = self._extension_default_fallback
        return model

    def update_value(self, value: Any) -> None:
        """Update the value of the widget."""
        if hasattr(self.widget, "update_value"):
            self.widget.update_value(value)
        else:
            model = self.to_model()
            self.update_model(model.with_value(value))

    def update_metadata(self, metadata: Any) -> None:
        """Update the metadata of the widget data model."""
        return self.update_model(self.to_model().with_metadata(metadata))

    def write_model(self, path: str | Path, plugin: str | None = None) -> None:
        """Write the widget data to a file."""
        return self._write_model(path, plugin, self.to_model())

    @property
    def tab_area(self) -> TabArea[_W]:
        """Tab area of the sub-window."""
        _hash = self._main_window()._tab_hash_for_window(self._frontend_widget())
        return self._main_window()._himena_main_window.tabs._tab_areas[_hash]

    def _write_model(
        self, path: str | Path, plugin: str | None, model: WidgetDataModel
    ) -> None:
        io_utils.write(model, path, plugin=plugin)
        self.update_default_save_path(path)
        return None

    def _set_modification_tracking(self, enabled: bool) -> None:
        if self._data_modifications.track_enabled == enabled:
            return None  # already in the desired state
        if enabled and self.supports_to_model:
            self._data_modifications = Modifications(
                initial_value=self.to_model().value,
                track_enabled=True,
            )
        else:
            self._data_modifications = Modifications(
                initial_value=None, track_enabled=False
            )

    def _set_state(self, value: WindowState, inst: BackendInstructions | None = None):
        main = self._main_window()
        front = self._frontend_widget()
        if main._window_state(front) is value:  # if already in the state, do nothing
            return None
        if inst is None:
            inst = main._himena_main_window._instructions
        main._set_window_state(front, value, inst)
        stack = self.tab_area._minimized_window_stack_layout
        if value is WindowState.MIN:
            stack.add(self)
            self.anchor = None
        elif value in (WindowState.FULL, WindowState.MAX):
            self.anchor = _anchor.AllCornersAnchor()
        else:
            self.anchor = None
            if self._parent_layout_ref() is stack:
                stack.remove(self)
                stack._reanchor(Size(*main._area_size()))

    def _set_rect(
        self,
        value: tuple[int, int, int, int],
        inst: BackendInstructions | None = None,
    ):
        if inst is None:
            inst = self._main_window()._himena_main_window._instructions
        rect_old = self.rect
        main = self._main_window()
        front = self._frontend_widget()
        rect = WindowRect.from_tuple(*value)
        anc = self.anchor.update_for_window_rect(main._area_size(), rect)
        main._set_window_rect(front, rect, inst)
        self.anchor = anc
        if parent := self._parent_layout_ref():
            parent._adjust_child_resize(self, rect_old, rect)

    def _reanchor(self, size: Size):
        if self.state is WindowState.MIN:
            pass
        elif self.state in (WindowState.MAX, WindowState.FULL):
            self.rect = WindowRect(0, 0, *size)
        else:
            super()._reanchor(size)

    def _process_drop_event(
        self,
        incoming: DragDataModel,
        source: SubWindow[_W] | None = None,
    ) -> bool:
        if hasattr(self.widget, "dropped_callback"):
            # to remember how the model was mapped to a widget class
            model = incoming.data_model()
            if source is not None:
                model.force_open_with = get_widget_class_id(source.widget)
            drop_result = self.widget.dropped_callback(model)
            if drop_result is None:
                drop_result = DropResult()
            ui = self._main_window()._himena_main_window
            if drop_result.command_id:
                # record the command
                out = ui.exec_action(
                    drop_result.command_id,
                    window_context=self,
                    model_context=self.to_model(),
                    with_params=drop_result.with_params,
                    process_model_output=False,
                )
                if isinstance(out, WidgetDataModel):
                    self.update_model(out)
                    self._update_model_workflow(out.workflow)
            if source is not None:
                if drop_result.delete_input:
                    source._close_me(ui)
                ui._backend_main_window._move_focus_to(source._frontend_widget())
            return True
        return False

    def update(
        self,
        *,
        rect: tuple[int, int, int, int] | WindowRect | None = None,
        state: WindowState | None = None,
        title: str | None = None,
        anchor: _anchor.WindowAnchor | str | None = None,
    ) -> SubWindow[_W]:
        """A helper method to update window properties."""
        if rect is not None:
            self.rect = rect
        if state is not None:
            self.state = state
        if title is not None:
            self.title = title
        if anchor is not None:
            self.anchor = anchor
        return self

    def add_child(
        self,
        widget: _W,
        *,
        title: str | None = None,
    ) -> SubWindow[_W]:
        """Add a child sub-window, which is automatically closed when the parent is closed."""  # noqa: E501
        main = self._main_window()._himena_main_window
        i_tab, _ = self._find_me(main)
        child = main.tabs[i_tab].add_widget(widget, title=title)
        self._child_windows.add(child)
        return child

    def _find_me(self, main: MainWindow) -> tuple[int, int]:
        for i_tab, tab in main.tabs.enumerate():
            for i_win, win in tab.enumerate():
                # NOTE: should not be `win is self`, because the wrapper may be
                # recreated
                if win.widget is self.widget:
                    return i_tab, i_win
        raise RuntimeError(f"SubWindow {self.title} not found in main window.")

    def _close_me(self, main: MainWindow, confirm: bool = False) -> None:
        if self._need_ask_save_before_close() and confirm:
            title_short = repr(self.title)
            if len(title_short) > 60:
                title_short = title_short[:60] + "..."
            if isinstance(self.save_behavior, SaveToNewPath):
                message = f"{title_short} is not saved yet. Save before closing?"
            else:
                message = f"Save changes to {title_short}?"
            resp = main.exec_choose_one_dialog(
                title="Closing window",
                message=message,
                choices=["Save", "Don't save", "Cancel"],
            )
            if resp is None or resp == "Cancel":
                return None
            elif resp == "Save":
                if cb := self._save_from_dialog(main):
                    cb()
                else:
                    return None

        i_tab, i_win = self._find_me(main)
        del main.tabs[i_tab][i_win]

    def _save_from_dialog(
        self,
        main: MainWindow,
        behavior: SaveBehavior | None = None,
        plugin: str | None = None,
    ) -> Callable[[], Path] | None:
        """Save this window to a new path, return if saved."""
        if behavior is None:
            behavior = self.save_behavior
        model = self.to_model()
        if save_path := behavior.get_save_path(main, model):

            def _save():
                main.set_status_tip(f"Saving {self.title!r} to {save_path}", duration=2)
                self._write_model(save_path, plugin=plugin, model=model)
                main.set_status_tip(f"Saved {self.title!r} to {save_path}", duration=2)
                return save_path

            return _save
        return None

    def _close_all_children(self, main: MainWindow) -> None:
        """Close all the sub-windows that are children of this window."""
        for child in self._child_windows:
            child._close_all_children(main)
            if child.is_alive:
                child._close_me(main, confirm=False)

    def _close_callback(self):
        main = self._main_window()._himena_main_window
        self._close_all_children(main)
        if layout := self._parent_layout_ref():
            layout.remove(self)
        self._alive = False
        main.events.window_closed.emit(self)

    def _determine_read_from(self) -> tuple[Path | list[Path], str | None] | None:
        """Determine how can the data be efficiently read."""
        workflow = self._widget_workflow.last()
        if isinstance(workflow, LocalReaderMethod):
            return workflow.path, workflow.plugin
        elif isinstance(save_bh := self.save_behavior, SaveToPath):
            return save_bh.path, None
        else:
            return None

    def _update_from_returned_model(self, model: WidgetDataModel) -> SubWindow[_W]:
        """Update the sub-window based on the returned model."""
        if (wf := model.workflow.last()) is not None:
            if isinstance(wf, LocalReaderMethod):
                # file is directly read from the local path
                if isinstance(save_path := wf.path, Path):
                    self.update_default_save_path(save_path, plugin=wf.plugin)
            elif isinstance(wf, CommandExecution):
                # model is created by some command
                if not isinstance(model.save_behavior_override, NoNeedToSave):
                    self._set_ask_save_before_close(True)
            self._identifier = wf.id
        if len(wlist := model.workflow) > 0:
            self._update_model_workflow(wlist)
        if save_behavior_override := model.save_behavior_override:
            self._save_behavior = save_behavior_override
        if not model.editable:
            with suppress(AttributeError):
                self.is_editable = False

        if model.is_subtype_of("text"):
            self._set_modification_tracking(True)
        return self

    def _switch_to_file_watch_mode(self):
        # TODO: don't use Qt in the future
        from himena.qt._qtwatchfiles import QWatchFileObject

        self.title = f"[Preview] {self.title}"
        QWatchFileObject(self)
        return None


class ParametricWindow(SubWindow[_W]):
    """Subwindow with a parametric widget inside."""

    _IS_PREVIEWING = "is_previewing"  # keyword argument used for preview flag
    btn_clicked = Signal(object)  # emit self
    params_changed = Signal(object)  # emit self

    def __init__(
        self,
        widget: _W,
        callback: Callable,
        main_window: BackendMainWindow[_W],
    ):
        super().__init__(widget, main_window)
        self._callback = callback
        self.btn_clicked.connect(self._widget_callback)
        self._preview_window_ref: Callable[[], WidgetWrapper[_W] | None] = _do_nothing
        self._auto_close = True
        self._run_asynchronously = False
        self._last_future: Future | None = None
        self._result_as: Literal["window", "below", "right"] = "window"

        # check if callback has "is_previewing" argument
        sig = inspect.signature(callback)
        self._has_is_previewing = self._IS_PREVIEWING in sig.parameters
        self._fn_signature = sig

    def get_params(self) -> dict[str, Any]:
        """Get the parameters of the widget."""
        if hasattr(self.widget, PWPN.GET_PARAMS):
            params = getattr(self.widget, PWPN.GET_PARAMS)()
            if not isinstance(params, dict):
                raise TypeError(
                    f"`{PWPN.GET_PARAMS}` of {self.widget!r} must return a dict, got "
                    f"{type(params)}."
                )
        else:
            params = {}
        return params

    def update_params(self, params: dict[str, Any], **kwargs) -> None:
        """Update the parameters of the widget."""
        if hasattr(self.widget, PWPN.UPDATE_PARAMS):
            getattr(self.widget, PWPN.UPDATE_PARAMS)({**params, **kwargs})
        else:
            raise NotImplementedError(
                f"{self.widget!r} does not support setting parameters."
            )
        self._emit_param_changed()

    def _get_preview_window(self) -> SubWindow[_W] | None:
        """Return the preview window if it is alive."""
        if (prev := self._preview_window_ref()) and prev.is_alive:
            return prev
        return None

    def _is_run_immediately(self) -> bool:
        for param in self._fn_signature.parameters.values():
            annot = param.annotation
            if _is_annotated(annot):
                _, op = _split_annotated_type(annot)
                if "bind" not in op:
                    return False
            else:
                return False
        return True

    def _widget_callback(self):
        """Callback when the call button is clicked."""
        main = self._main_window()
        main._set_parametric_widget_busy(self, True)
        try:
            self._callback_with_params(self.get_params())
        except Exception:
            main._set_parametric_widget_busy(self, False)
            raise

    def _call(self, **kwargs):
        """Call the callback (maybe) asynchronously."""
        ui = self._main_window()._himena_main_window
        if self._run_asynchronously:
            if self._last_future is not None:
                self._last_future.cancel()
                self._last_future = None
            self._last_future = future = ui._executor.submit(self._callback, **kwargs)
            return future
        else:
            return self._callback(**kwargs)

    def _widget_preview_callback(self):
        """Callback function of parameter change during preview"""
        main = self._main_window()
        if not self.is_preview_enabled():
            if prev := self._get_preview_window():
                self._preview_window_ref = _do_nothing
                self._child_windows.discard(prev)
                if self._result_as == "window":
                    prev._close_me(main._himena_main_window)
                else:
                    main._remove_widget_from_parametric_window(self)
                    if hint := self.size_hint():
                        self.rect = (self.rect.left, self.rect.top, hint[0], hint[1])
            return None
        kwargs = self.get_params()
        if self._has_is_previewing:
            kwargs[self._IS_PREVIEWING] = True
        return_value = self._call(**kwargs)
        if isinstance(return_value, Future):  # running asynchronously
            done = main._process_future_done_callback(
                self._widget_preview_callback_done,
                lambda e: main._set_parametric_widget_busy(self, False),
            )
            return_value.add_done_callback(done)
            main._set_parametric_widget_busy(self, True)
        else:
            temp_future = Future()
            temp_future.set_result(return_value)
            self._widget_preview_callback_done(temp_future)

    def _widget_preview_callback_done(self, future: Future):
        """Callback function when the job of preview is done."""
        main = self._main_window()
        main._set_parametric_widget_busy(self, False)
        return_value = future.result()
        if return_value is None:
            return None
        if not isinstance(return_value, WidgetDataModel):
            raise NotImplementedError(
                "Preview is only supported for WidgetDataModel but the return value "
                f"was {type(return_value)}"
            )
        if prev := self._get_preview_window():
            prev.update_model(return_value)
        else:
            # create a new preview window
            result_widget = self._model_to_new_window(return_value)
            if self._result_as == "window":
                # create a new preview window
                title = f"{return_value.title} (preview)"
                prev = self.add_child(result_widget, title=title)
                main.set_widget_as_preview(prev)
                prev.force_not_editable(True)  # disable editing if possible
                # move the window so that it does not overlap with the parametric window
                prev.rect = prevent_window_overlap(self, prev, main._area_size())
            else:
                main._add_widget_to_parametric_window(
                    self, result_widget, self._result_as
                )
                # update the size because new window is added
                if hint := self.size_hint():
                    self.rect = (self.rect.left, self.rect.top, hint[0], hint[1])
                prev = WidgetWrapper(result_widget, main)  # just for wrapping
            self._preview_window_ref = weakref.ref(prev)
            main._move_focus_to(self._frontend_widget())
        return None

    def _process_return_value(self, return_value: Any, kwargs: dict[str, Any]):
        main = self._main_window()
        ui = main._himena_main_window
        main._set_parametric_widget_busy(self, False)
        tracker = ModelTrack.get(self._callback)
        _return_annot = self._fn_signature.return_annotation
        _LOGGER.info("Got tracker: %r", tracker)
        if isinstance(return_value, WidgetDataModel):
            if prev := self._get_preview_window():
                # no need to create a new window, just use the preview window
                self._preview_window_ref = _do_nothing
                if self._result_as != "window":
                    widget = prev.widget  # avoid garbage collection
                    main._remove_widget_from_parametric_window(self)
                    result_widget = ui.add_widget(widget)
                    result_widget._update_from_returned_model(return_value)
                else:
                    self._child_windows.discard(prev)
                    result_widget = prev
                result_widget.title = return_value.title  # title needs update

                # if callback has "is_previewing" argument, the returned value may
                # differ, thus the widget needs update.
                if self._has_is_previewing:
                    result_widget.update_model(return_value)
                result_widget.force_not_editable(False)
                with suppress(AttributeError):
                    result_widget.is_editable = True
                if self._auto_close:
                    self._close_me(ui)
            else:
                result_widget = self._process_model_output(return_value, tracker)
                if result_widget is None:
                    if tracker is not None:
                        new_workflow = tracker.to_workflow(kwargs)
                        return_value.workflow = new_workflow  # needs inheritance
                    return None
                elif return_value.update_inplace:
                    result_widget._widget_workflow = tracker.to_workflow(kwargs)
            _LOGGER.info("Got subwindow: %r", result_widget)
            if tracker is not None:
                new_workflow = tracker.to_workflow(kwargs)
                _LOGGER.info(
                    "Inherited method %r, where the original method was %r",
                    new_workflow,
                    return_value.workflow,
                )
                # NOTE: overwrite=False is needed to avoid overwriting ReaderMethod
                result_widget._update_model_workflow(new_workflow, overwrite=False)
                if isinstance(new_workflow, CommandExecution):
                    if not isinstance(
                        return_value.save_behavior_override, NoNeedToSave
                    ):
                        result_widget._set_ask_save_before_close(True)
            ui.events.window_added.emit(result_widget)
        elif _return_annot in (Parametric, ParametricWidgetProtocol):
            raise NotImplementedError
        else:
            annot = getattr(self._callback, "__annotations__", {})
            if isinstance(return_value, Future):
                injection_type_hint = Future
                # This is hacky. The injection store will process the result but the
                # return type cannot be inherited from the callback. Here, we just set
                # the type hint to Future and let it processed in the
                # "_future_done_callback" method of himena application.
                if prev := self._get_preview_window():
                    top_left = (prev.rect.left, prev.rect.top)
                    size = prev.rect.size()
                else:
                    top_left = (self.rect.left, self.rect.top)
                    size = None
                injection_ns = ui.model_app.injection_store.namespace
                FutureInfo(
                    type_hint=annot.get("return", None),
                    track=tracker,
                    kwargs=kwargs,
                    top_left=top_left,
                    size=size,
                ).resolve_type_hint(injection_ns).set(return_value)
            else:
                injection_type_hint = annot.get("return", None)
            self._process_other_output(return_value, injection_type_hint)
        return None

    def _callback_with_params(
        self,
        kwargs: dict[str, Any],
        force_sync: bool = False,
        force_close: bool = False,
    ) -> Any:
        if self._has_is_previewing:
            kwargs = {**kwargs, self._IS_PREVIEWING: False}
        main = self._main_window()
        old_run_async = self._run_asynchronously
        try:
            if force_sync:
                self._run_asynchronously = False
            return_value = self._call(**kwargs)
        except Exception:
            main._set_parametric_widget_busy(self, False)
            raise
        finally:
            self._run_asynchronously = old_run_async
        if isinstance(return_value, Future):
            main._add_job_progress(return_value, desc=self.title, total=0)
            return_value.add_done_callback(
                main._process_future_done_callback(
                    self._process_return_value,
                    lambda e: main._set_parametric_widget_busy(self, False),
                    kwargs=kwargs,
                )
            )
        else:
            main._set_parametric_widget_busy(self, False)
            self._process_return_value(return_value, kwargs)
        if (not self._auto_close) and force_close:
            self._close_me(main._himena_main_window)
        return return_value

    def is_preview_enabled(self) -> bool:
        """Whether the widget supports preview."""
        isfunc = getattr(self.widget, PWPN.IS_PREVIEW_ENABLED, None)
        return callable(isfunc) and isfunc()

    def _emit_btn_clicked(self) -> None:
        return self.btn_clicked.emit(self)

    def _emit_param_changed(self) -> None:
        return self.params_changed.emit(self)

    def _process_model_output(
        self,
        model: WidgetDataModel,
        tracker: ModelTrack | None = None,
    ) -> SubWindow[_W] | None:
        """Process the returned WidgetDataModel."""
        ui = self._main_window()._himena_main_window
        i_tab, i_win = self._find_me(ui)
        rect = self.rect
        if self._auto_close:
            del ui.tabs[i_tab][i_win]
        if model.update_inplace and tracker and tracker.contexts:
            if win := ui._window_for_workflow_id(tracker.contexts[0].value):
                win.update_model(model)
                return win
        if ui._instructions.process_model_output:
            widget = self._model_to_new_window(model)
            result_widget = ui.tabs[i_tab].add_widget(
                widget, title=model.title, auto_size=False
            )
            # coerce rect
            if size_hint := result_widget.size_hint():
                new_rect = (rect.left, rect.top, size_hint[0], size_hint[1])
            else:
                new_rect = rect
            result_widget.rect = new_rect
            _checker.call_widget_added_callback(widget)
            return result_widget._update_from_returned_model(model)
        return None

    def _model_to_new_window(self, model: WidgetDataModel) -> _W:
        ui = self._main_window()._himena_main_window
        widget = ui._pick_widget(model)
        return widget

    def _process_other_output(self, return_value: Any, type_hint: Any | None = None):
        _LOGGER.info("Got output: %r with type hint %r", type(return_value), type_hint)
        ui = self._main_window()._himena_main_window
        ui.model_app.injection_store.process(return_value, type_hint=type_hint)
        if self._auto_close:
            with suppress(RuntimeError):
                # FIXME: if the async command does not require parameter input, this
                # window is already closed. We just ignore the error for now.
                self._close_me(ui)


class DockWidget(WidgetWrapper[_W]):
    """Dock widget wrapper."""

    def __init__(
        self,
        widget: _W,
        main_window: BackendMainWindow[_W],
        identifier: uuid.UUID | None = None,
    ):
        super().__init__(widget, main_window)
        if identifier is not None:
            self._identifier = identifier
        self._has_update_configs = hasattr(widget, "update_configs")
        self._parse_config_cache: Callable[[dict], Any] | None = None
        self._command_id: str | None = None

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(title={self.title!r}, widget={_widget_repr(self)})"
        )

    @property
    def visible(self) -> bool:
        """Visibility of the dock widget."""
        return self._main_window()._dock_widget_visible(self._frontend_widget())

    @visible.setter
    def visible(self, visible: bool) -> bool:
        return self._main_window()._set_dock_widget_visible(
            self._frontend_widget(), visible
        )

    def show(self) -> None:
        """Show the dock widget."""
        self.visible = True

    def hide(self) -> None:
        """Hide the dock widget."""
        self.visible = False

    @property
    def title(self) -> str:
        """Title of the dock widget."""
        return self._main_window()._dock_widget_title(self._frontend_widget())

    @title.setter
    def title(self, title: str) -> None:
        return self._main_window()._set_dock_widget_title(
            self._frontend_widget(), str(title)
        )

    def _parse_config(self, cfg_dict: dict[str, Any]) -> Any:
        if self._parse_config_cache is not None:
            return self._parse_config_cache(**cfg_dict)
        cfgs = AppActionRegistry.instance()._plugin_default_configs
        cfg_type = cfgs[self._command_id].config_class
        self._parse_config_cache = cfg_type
        return cfg_type(**cfg_dict)

    def update_configs(self, cfg: Any):
        """Update the configuration of the dock widget."""
        if self._has_update_configs:
            if isinstance(cfg, dict):
                cfg = self._parse_config(cfg)
            self.widget.update_configs(cfg)
        return None


def _widget_repr(wrapper: WidgetWrapper[_W]) -> str:
    if widget := wrapper._widget():
        wid = get_widget_class_id(type(widget))
    else:
        wid = "deleted"
    return f"<{wid}>"


def _do_nothing() -> None:
    return None



from contextlib import contextmanager
import tempfile
from typing import Iterator, Literal, Any, TYPE_CHECKING
from pathlib import Path
import subprocess

from pydantic import Field
from himena.consts import StandardType
from himena.exceptions import NotExecutable
from himena.utils.misc import PluginInfo
from himena.utils.cli import remote_to_local, wsl_to_local
from himena.workflow._base import WorkflowStep

if TYPE_CHECKING:
    from himena.types import WidgetDataModel
    from himena.workflow import Workflow


class NoParentWorkflow(WorkflowStep):
    """Describes that one has no parent."""

    output_model_type: str | None = Field(default=None)

    def iter_parents(self) -> Iterator[int]:
        yield from ()


class ProgrammaticMethod(NoParentWorkflow):
    """Describes that one was created programmatically."""

    type: Literal["programmatic"] = "programmatic"

    def _get_model_impl(self, wf):
        raise NotExecutable("Data was added programmatically, thus cannot be executed.")


class UserInput(NoParentWorkflow):
    """Describes that one will be provided as a runtime user input."""

    type: Literal["input"] = "input"
    label: str = Field(default="")
    doc: str = Field(default="")
    how: Literal["file", "model"] = "model"

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        raise NotExecutable("No user input bound to this workflow step.")


class RuntimeInputBound(UserInput):
    bound_value: Any

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        return self.bound_value


class ReaderMethod(NoParentWorkflow):
    """Describes that one was read from a file."""

    plugin: str | None = Field(default=None)
    metadata_override: Any | None = Field(default=None)

    def run(self) -> "WidgetDataModel":
        raise NotImplementedError

    def _run_store(self, path: Path | list[Path]) -> "WidgetDataModel[Any]":
        from himena._providers import ReaderStore

        store = ReaderStore.instance()
        return store.run(path, plugin=self.plugin)


class PathReaderMethod(ReaderMethod):
    """Describes that one was read from a path."""

    path: Path | list[Path]
    force_directory: bool = Field(default=False)

    def run_command(self, dst_path: Path, stdout=None):
        """Run command to move the file from source to local `dst_path`."""
        raise NotImplementedError("No conversion command defined for this method.")

    def to_str(self) -> str:
        """Return the local file path representation."""
        if isinstance(self.path, Path):
            return self.path.as_posix()
        return ";".join(p.as_posix() for p in self.path)

    def _list_dst_filenames(self) -> list[str]:
        if isinstance(self.path, Path):
            filenames = [self.path.name]
        else:
            filenames: list[Path] = []
            for p in self.path:
                filenames.append(p.name)
        return filenames

    @contextmanager
    def run_context(
        self, filenames: list[str] | None = None
    ) -> Iterator[Path | list[Path]]:
        if filenames is None:
            filenames = self._list_dst_filenames()
        with tempfile.TemporaryDirectory() as tmpdir:
            if isinstance(self.path, Path):
                dst_path = Path(tmpdir, filenames[0])
                self.run_command(dst_path)
            else:
                dst_path: list[Path] = []
                for p in filenames:
                    dst_p = Path(tmpdir, p)
                    self.run_command(dst_p)
                    dst_path.append(dst_p)
            yield dst_path

    def run(self) -> "WidgetDataModel":
        from himena._providers import ReaderStore
        from himena.types import WidgetDataModel

        filenames = self._list_dst_filenames()
        store = ReaderStore.instance()
        matched = store.get(Path(filenames[0]), empty_ok=True, min_priority=1)
        if len(filenames) > 1 or matched:
            with self.run_context(filenames) as dst_path:
                model = self._run_store(dst_path)
        else:
            # If the file type is not supported, no need to rsync it.
            model = WidgetDataModel(
                value=self.path,
                type=StandardType.READER_NOT_FOUND,
            )

        model.title = self.path.name
        model.workflow = self.construct_workflow()
        return model


class LocalReaderMethod(ReaderMethod):
    """Describes that one was read from a local source file."""

    type: Literal["local-reader"] = "local-reader"
    path: Path | list[Path]

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel[Any]":
        out = self.run()
        if main := wf._mock_main_window:
            win = main.add_data_model(out)
            win._identifier = self.id
        return out

    @property
    def force_directory(self) -> bool:
        # just for compatibility with PathReaderMethod
        return isinstance(self.path, Path) and self.path.is_dir()

    def to_str(self) -> str:
        """Return the local file path representation."""
        if isinstance(self.path, Path):
            return self.path.as_posix()
        return ";".join(p.as_posix() for p in self.path)

    def run(self) -> "WidgetDataModel[Any]":
        """Get model by importing the reader plugin and actually read the file(s)."""
        from himena.types import WidgetDataModel

        model = self._run_store(self.path)
        if not isinstance(model, WidgetDataModel):
            raise ValueError(f"Expected to return a WidgetDataModel but got {model}")
        if len(model.workflow) == 0:
            model = model._with_source(
                source=self.path,
                plugin=PluginInfo.from_str(self.plugin) if self.plugin else None,
                id=self.id,
            )
        if self.metadata_override is not None:
            model.metadata = self.metadata_override
        return model


class RemoteReaderMethod(PathReaderMethod):
    """Describes that one was read from a remote source file."""

    type: Literal["remote-reader"] = "remote-reader"
    host: str
    username: str
    port: int = Field(default=22)
    wsl: bool = Field(default=False)
    protocol: str = Field(default="rsync")

    @classmethod
    def from_str(
        cls,
        s: str,
        /,
        wsl: bool = False,
        protocol: str = "rsync",
        output_model_type: str | None = None,
        force_directory: bool = False,
    ) -> "RemoteReaderMethod":
        username, rest = s.split("@")
        host, path = rest.split(":")
        if ";" in path:
            path = [Path(p) for p in path.split(";")]
        else:
            path = Path(path)
        return cls(
            username=username,
            host=host,
            path=path,
            wsl=wsl,
            protocol=protocol,
            output_model_type=output_model_type,
            force_directory=force_directory,
        )

    def to_str(self) -> str:
        """Return the remote file path representation."""
        if isinstance(self.path, Path):
            return f"{self.username}@{self.host}:{self.path.as_posix()}"
        return (
            f"{self.username}@{self.host}:{';'.join(p.as_posix() for p in self.path)}"
        )

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        out = self.run()
        if main := wf._mock_main_window:
            win = main.add_data_model(out)
            win._identifier = self.id
        return out

    def run_command(self, dst_path: Path, stdout=None):
        """Run scp/rsync command to move the file from remote to local `dst_path`."""
        if isinstance(self.path, Path):
            args = remote_to_local(
                self.protocol,
                self.to_str(),
                dst_path,
                is_wsl=self.wsl,
                is_dir=self.force_directory,
                port=self.port,
            )
        else:
            raise ValueError(
                "Cannot run command for multiple paths in RemoteReaderMethod."
            )
        result = subprocess.run(args, stdout=stdout)
        if result.returncode != 0:
            raise ValueError(f"Failed to run command {args}: {result!r}")


class WslReaderMethod(PathReaderMethod):
    """Describes that one was read from a WSL source file."""

    type: Literal["wsl-reader"] = "wsl-reader"

    @classmethod
    def from_str(cls, line: str, force_directory: bool = False) -> "WslReaderMethod":
        """Construct a WslReaderMethod from a string representation."""
        if ";" in line:
            path = [Path(p) for p in line.split(";")]
        else:
            path = Path(line)
        return cls(
            path=path,
            force_directory=force_directory,
        )

    def run_command(self, dst_path: Path, stdout=None):
        """Run cp command to move the file from wsl to local `dst_path`."""

        if isinstance(self.path, Path):
            args = wsl_to_local(
                self.path.as_posix(), dst_path, is_dir=self.force_directory
            )
            result = subprocess.run(args, stdout=stdout)
        else:
            raise ValueError(
                "Cannot run command for multiple paths in WSLReaderMethod."
            )
        if result.returncode != 0:
            raise ValueError(f"Failed to run command {args}: {result!r}")



import json
from pathlib import Path
from typing import TYPE_CHECKING
from pydantic import BaseModel
from himena.utils.misc import iter_subclasses

if TYPE_CHECKING:
    from typing import Self

_META_NAME = "meta.json"
_CLASS_JSON = ".class.json"


class BaseMetadata(BaseModel):
    """The base class for a model metadata."""

    @classmethod
    def from_metadata(cls, dir_path: Path) -> "Self":
        """Construct the metadata from a directory."""
        return cls.model_validate_json(dir_path.joinpath(_META_NAME).read_text())

    def write_metadata(self, dir_path: Path) -> None:
        """Write the metadata to a directory."""
        dir_path.joinpath(_META_NAME).write_text(self.model_dump_json())

    def expected_type(self) -> str | None:
        """Return the expected type of the metadata. None if not applicable."""
        return None

    def _class_info(self) -> dict:
        return {"name": self.__class__.__name__, "module": self.__class__.__module__}

    def _repr_pretty_(self, p, cycle):
        """Pretty print the metadata."""
        lines = [f"{self.__class__.__name__}("]
        for key, value in self.__repr_args__():
            lines.append(f"  {key}={value!r},")
        lines.append(")")
        p.text("\n".join(lines))


def read_metadata(dir_path: Path) -> BaseMetadata:
    """Read the metadata from a directory."""
    with dir_path.joinpath(_CLASS_JSON).open("r") as f:
        class_js = json.load(f)
    module = class_js["module"]
    name = class_js["name"]
    for sub in iter_subclasses(BaseMetadata):
        if sub.__name__ == name and sub.__module__ == module:
            metadata_class = sub
            break
    else:
        raise ValueError(f"Metadata class {name=}n {module=} not found.")

    return metadata_class.from_metadata(dir_path)


def write_metadata(meta: BaseMetadata, dir_path: Path) -> None:
    """Write the metadata to a directory."""
    meta.write_metadata(dir_path)
    dir_path.joinpath(_CLASS_JSON).write_text(json.dumps(meta._class_info(), indent=4))
    return None



from dataclasses import field, MISSING
from typing import Any


def config_field(
    default=MISSING,
    *,
    default_factory=MISSING,
    tooltip: str | None = None,
    label: str | None = None,
    choices: list[str] | list[tuple[str, Any]] | None = None,
    widget_type: Any | None = None,
    enabled: bool | None = None,
    visible: bool | None = None,
    **kwargs,
):
    """Field used for plugin config dataclass.

    ```python
    @dataclass
    class MyPluginConfig:
        my_field: str = config_field("abc", tooltip="how to use this")
    ```
    """
    metadata = kwargs.copy()
    if tooltip is not None:
        metadata["tooltip"] = tooltip
    if label is not None:
        metadata["label"] = label
    if choices is not None:
        metadata["choices"] = choices
    if enabled is not None:
        metadata["enabled"] = enabled
    if visible is not None:
        metadata["visible"] = visible
    if widget_type is not None:
        metadata["widget_type"] = widget_type

    return field(
        default=default,
        default_factory=default_factory,
        compare=False,
        metadata=metadata,
    )



from __future__ import annotations

import re
from typing import Callable
from magicgui.types import Undefined
from magicgui.widgets import LineEdit, PushButton, Label
from magicgui.widgets.bases import ValuedContainerWidget

SLICE_PATTERN = re.compile(r"^\s*\d*\s*:\s*\d*\s*,\s*\d*\s*:\s*\d*\s*$")

SelectionType = tuple[tuple[int, int], tuple[int, int]]


class SelectionEdit(ValuedContainerWidget[SelectionType]):
    def __init__(
        self,
        value=Undefined,
        bind=Undefined,
        getter: Callable[[], SelectionType] | None = None,
        **kwargs,
    ):
        self._label_0 = Label(value="A[")
        self._line_edit = LineEdit()
        self._label_1 = Label(value="]")
        self._read_btn = PushButton(text="Read Selection")
        self._selection_getter = getter
        self._read_btn.clicked.connect(self._get_selection)
        if getter is None:
            self._read_btn.enabled = False
        super().__init__(
            value=value,
            bind=bind,
            widgets=[self._label_0, self._line_edit, self._label_1, self._read_btn],
            layout="horizontal",
            labels=False,
            **kwargs,
        )
        self.margins = (0, 0, 0, 0)

    def _get_selection(self):
        self.set_value(self._selection_getter(self))

    def get_value(self) -> SelectionType | None:
        text = self._line_edit.value.strip()
        if text == "":
            return None
        if not SLICE_PATTERN.match(text):
            raise ValueError(f"Cannot interpret {text} as slices")
        rsl_str, csl_str = text.split(",")
        r0, r1 = rsl_str.split(":")
        c0, c1 = csl_str.split(":")
        return (
            (_str_to_index(r0), _str_to_index(r1)),
            (_str_to_index(c0), _str_to_index(c1)),
        )

    def set_value(self, value: SelectionType | None):
        if value is None:
            self._line_edit.value = ""
        else:
            rsl, csl = value
            r0, r1 = rsl
            c0, c1 = csl
            r0 = _index_to_str(r0)
            r1 = _index_to_str(r1)
            c0 = _index_to_str(c0)
            c1 = _index_to_str(c1)
            self._line_edit.value = f"{r0}:{r1}, {c0}:{c1}"


def _index_to_str(s: int | None) -> str:
    return "" if s is None else s


def _str_to_index(s: str) -> int | None:
    s0 = s.strip()
    if s0 == "":
        return None
    return int(s0)



from typing import TYPE_CHECKING
from pathlib import Path
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from himena.widgets import MainWindow
    from himena.types import WidgetDataModel


class SaveBehavior(BaseModel):
    """A class that describes how a widget should be saved."""

    def get_save_path(
        self,
        main: "MainWindow",
        model: "WidgetDataModel",
    ) -> Path | None:
        """Return the path to save (None to cancel)."""
        return main.exec_file_dialog(
            mode="w",
            extension_default=model.extension_default,
            allowed_extensions=model.extensions,
            start_path=self._determine_save_path(model),
        )

    @staticmethod
    def _determine_save_path(model: "WidgetDataModel") -> str | None:
        if model.title is None:
            if model.extension_default is None:
                start_path = None
            else:
                start_path = f"Untitled{model.extension_default}"
        else:
            if Path(model.title).suffix in model.extensions:
                start_path = model.title
            elif model.extension_default is not None:
                start_path = Path(model.title).stem + model.extension_default
            else:
                start_path = model.title
        return start_path


class NoNeedToSave(SaveBehavior):
    """Describes that the widget does not need to be saved.

    This save behavior is usually used for commands that create a new data. Users will
    not be asked to save the data when they close the window, but will be asked if the
    data is modified.
    """


class CannotSave(SaveBehavior):
    """Describes that the widget cannot be saved."""

    reason: str

    def get_save_path(self, main, model):
        raise ValueError(f"Cannot save this widget: {self.reason}")


class SaveToNewPath(SaveBehavior):
    """Describes that the widget should be saved to a new path."""


class SaveToPath(SaveBehavior):
    """Describes that the widget should be saved to a specific path.

    A subwindow that has been saved once should always be tagged with this behavior.
    """

    path: Path
    ask_overwrite: bool = Field(
        default=True,
        description="Ask before overwriting the file if `path` already exists.",
    )
    plugin: str | None = Field(
        default=None,
        description="The plugin to use if the file is read back.",
    )

    def get_save_path(
        self,
        main: "MainWindow",
        model: "WidgetDataModel",
    ) -> Path | None:
        if self.path.exists() and self.ask_overwrite:
            if main._instructions.confirm:
                res = main.exec_choose_one_dialog(
                    title="Overwrite?",
                    message=f"{self.path}\nalready exists, overwrite?",
                    choices=["Overwrite", "Select another path", "Cancel"],
                )
            else:
                res = "Overwrite"
            if res == "Cancel":
                return None
            elif res == "Select another path":
                if path := SaveToNewPath().get_save_path(main, model):
                    self.path = path
                else:
                    return None
            # If overwrite is allowed, don't ask again.
            self.ask_overwrite = False
        return self.path



from __future__ import annotations
from typing import Any, TypeVar

from magicgui.widgets.bases import ValueWidget
from magicgui.types import Undefined
from magicgui.application import use_app
from magicgui.backends._qtpy.widgets import QBaseValueWidget
from himena.qt._qmodeldrop import QModelDrop, QModelDropList
from himena.types import WidgetDataModel
from himena.widgets import SubWindow

_W = TypeVar("_W")


class QMagicguiModelDrop(QBaseValueWidget):
    _qwidget: QModelDrop

    def __init__(self, **kwargs: Any) -> None:
        types = kwargs.get("types", None)
        layout = kwargs.get("layout", "horizontal")
        super().__init__(
            lambda parent: QModelDrop(types=types, layout=layout, parent=parent),
            "to_model",
            "set_model",
            "valueChanged",
            **kwargs,
        )


class QMagicguiSubWindowDrop(QBaseValueWidget):
    _qwidget: QModelDrop

    def __init__(self, **kwargs: Any) -> None:
        types = kwargs.get("types", None)
        layout = kwargs.get("layout", "horizontal")
        super().__init__(
            lambda parent: QModelDrop(types=types, layout=layout, parent=parent),
            "subwindow",
            "set_subwindow",
            "windowChanged",
            **kwargs,
        )


class QMagicguiModelDropList(QBaseValueWidget):
    _qwidget: QModelDropList

    def __init__(self, **kwargs: Any) -> None:
        types = kwargs.get("types", None)
        layout = kwargs.get("layout", "vertical")
        super().__init__(
            lambda parent: QModelDropList(types=types, layout=layout, parent=parent),
            "models",
            "set_models",
            "modelsChanged",
            **kwargs,
        )


class QMagicguiSubWindowDropList(QBaseValueWidget):
    _qwidget: QModelDropList

    def __init__(self, **kwargs: Any) -> None:
        types = kwargs.get("types", None)
        layout = kwargs.get("layout", "vertical")
        super().__init__(
            lambda parent: QModelDropList(types=types, layout=layout, parent=parent),
            "windows",
            "set_windows",
            "windowsChanged",
            **kwargs,
        )


class _ModelDropValueWidget(ValueWidget[_W]):
    def __init__(self, widget_type, value=Undefined, **kwargs):
        app = use_app()
        assert app.native

        if types := kwargs.pop("types", None):
            if isinstance(types, (list, tuple)):
                for t in types:
                    _assert_str(t)
            else:
                types = [_assert_str(types)]
        backend_kwargs = {"types": types}
        if layout := kwargs.pop("layout", None):
            backend_kwargs["layout"] = layout
        ValueWidget.__init__(
            self,
            value=value,
            widget_type=widget_type,
            backend_kwargs=backend_kwargs,
            **kwargs,
        )


class ModelDrop(_ModelDropValueWidget[WidgetDataModel]):
    def __init__(self, value=Undefined, **kwargs):
        super().__init__(QMagicguiModelDrop, value=value, **kwargs)

    def get_value(self) -> WidgetDataModel:
        out = super().get_value()
        if out is None and not self._nullable:
            raise ValueError(f"No model is specified to {self.label!r}.")
        return out


class SubWindowDrop(_ModelDropValueWidget[SubWindow]):
    def __init__(self, value=Undefined, **kwargs):
        super().__init__(QMagicguiSubWindowDrop, value=value, **kwargs)

    def get_value(self) -> SubWindow:
        out = super().get_value()
        if out is None and not self._nullable:
            raise ValueError(f"No model is specified to {self.label!r}.")
        return out


class ModelListDrop(_ModelDropValueWidget[list[WidgetDataModel]]):
    def __init__(self, value=Undefined, **kwargs):
        super().__init__(QMagicguiModelDropList, value=value, **kwargs)


class SubWindowListDrop(_ModelDropValueWidget[list[SubWindow]]):
    def __init__(self, value=Undefined, **kwargs):
        super().__init__(QMagicguiSubWindowDropList, value=value, **kwargs)


def _assert_str(t):
    if not isinstance(t, str):
        raise TypeError(f"types must be a str or a list of str, got {t}")
    return t



from pathlib import Path

import numpy as np
from himena.data_wrappers import wrap_dataframe, read_csv
import pytest

@pytest.mark.parametrize(
    "mod",
    ["dict", "pandas", "polars", "pyarrow"]
)
def test_read_csv(mod: str, sample_dir: Path, tmpdir):
    df_orig = read_csv(mod, sample_dir / "table.csv")
    df = wrap_dataframe(df_orig)
    repr(df.copy())
    assert df.column_names() == ["a", "b", "c"]
    assert df.column_to_array("a").tolist() == [1, 3, 4, 6, 8]
    assert df.dtypes[0].kind == "i"
    assert df.dtypes[1].kind == "f"
    df_filt = df.filter(np.array([True, False, True, False, True]))
    assert df_filt.shape == (3, 3)
    df.sort("b")
    df.sort("a", descending=True)

    df_cycled_csv = df.from_csv_string(df.to_csv_string())
    assert type(df_cycled_csv) is type(df)
    assert df_cycled_csv.column_names() == ["a", "b", "c"]
    assert df_cycled_csv.column_to_array("a").tolist() == [1, 3, 4, 6, 8]

    df_cycled_dict = df.from_dict(df.to_dict())
    assert df_cycled_dict.column_names() == ["a", "b", "c"]
    assert df_cycled_dict.column_to_array("a").tolist() == [1, 3, 4, 6, 8]

    assert df[0, 0] == 1
    assert df[1:3, 0].tolist() == [3, 4]
    assert df["a"].tolist() == [1, 3, 4, 6, 8]
    df.to_list()

    df_new = df.with_columns({"new": [1, 2, 3, 4, 5]})
    assert df_new.column_names() == ["a", "b", "c", "new"]

    save_dir = Path(tmpdir)
    df.write(save_dir / "table.csv")
    df.write(save_dir / "table.txt")
    df.write(save_dir / "table.tsv")
